using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMAtmosphere - Ambience v2.0
//
// A library of ambiant sounds

namespace VAMAAmbiencePlugin
{
    public class VAMAAmbience : MVRScript
    {	
        public static string PLUGIN_PATH;


		public Transform CUARoot;
		public Transform AmbienceRoot;
		private GameObject GoADebugMaxDist;
		public AudioSource AsAmbience;
		public AudioHighPassFilter AhpfAmbience;
		public AudioLowPassFilter AlpfAmbience;
		
		private bool cuaInitialized = false;
		
		// VAM UI	
		public JSONStorableStringChooser Ambience;
		protected List<string> ambienceChoices;
		
		public JSONStorableFloat AmbienceMaxDist;
		public JSONStorableFloat AmbienceVolume;
		public JSONStorableFloat AmbiencePitch;
		public JSONStorableFloat AmbienceSpatialization;
		public JSONStorableFloat AmbienceReverb;
		public JSONStorableBool AmbienceEnableOcclusion;
		public JSONStorableFloat AmbienceHighPass;
		public JSONStorableFloat AmbienceLowPass;
		
		public JSONStorableBool ShowDebug;
		public JSONStorableBool ShowOcclusionDebug;

		public JSONStorableFloat OcclusionLowestFrequency;
		public JSONStorableFloat OcclusionHighestFrequency;
		public JSONStorableFloat OcclusionVolumeAttenuation;
		public JSONStorableFloat OcclusionInterpolationSpeed;
		public JSONStorableFloat SoundWideness;
		public JSONStorableFloat ListenerWideness;
		
		// Occlusion variables
		private Transform CLTransform; // Current Listener Transform
		private LineRenderer[] _debugLinesList;
		private int debugLineCount = 0;
		private static int maxDebugLines = 9;
		private int raycastTotal = 0;
		private int raycastHitCount = 0;
		private float newCutoff = 0f;
		private float newVolumeAtt = 0f;
		
		// Security
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		
		// Assetbundle
		private Request audioBundleRequest = null;
		
		// Others
		private List<string> ambiencesPaths;
		private UIDynamic[] OcclusionUIElements;
		
			
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add VAMAAmbience on a Custom Unity Asset Atom");
                    return;
                }
				
				PLUGIN_PATH = GetPluginPath(this);		
				UIDynamicSlider tmpSlider;
				OcclusionUIElements = new UIDynamic[6];

				// Ambiences Choises
				ambienceChoices = new List<string>();
				
				Ambience = new JSONStorableStringChooser("Ambience", ambienceChoices, "", "Ambience", AmbienceCallback);
				UIDynamicPopup FAdp = CreateScrollablePopup(Ambience, false);
				FAdp.labelWidth = 180f;
				
				AmbienceMaxDist = new JSONStorableFloat("Ambience max distance", 25f, AmbienceMaxDistCallback, 1f, 100f);
				CreateSlider(AmbienceMaxDist, false);
				
				AmbienceVolume = new JSONStorableFloat("Ambience volume", 0.75f, AmbienceVolumeCallback, 0f, 1f);
				CreateSlider(AmbienceVolume, false);

				AmbiencePitch = new JSONStorableFloat("Ambience pitch", 1f, AmbiencePitchCallback, -3f, 3f);
				CreateSlider(AmbiencePitch, false);
				
				AmbienceSpatialization = new JSONStorableFloat("Ambience spatialization", 1f, AmbienceSpatializationCallback, 0f, 1f);
				CreateSlider(AmbienceSpatialization, false);
				
				AmbienceReverb = new JSONStorableFloat("Ambience reverb", 1f, AmbienceReverbCallback, 0f, 1.1f);
				CreateSlider(AmbienceReverb, false);
				
				AmbienceHighPass = new JSONStorableFloat("Ambience high pass", 10f, (val) => { AmbienceHighPass.valNoCallback = Mathf.Round(val); AmbienceHighPassCallback(); } , 10f, 22000f);
				CreateSlider(AmbienceHighPass, false);
				
				AmbienceLowPass = new JSONStorableFloat("Ambience low pass", 22000f, (val) => { AmbienceLowPass.valNoCallback = Mathf.Round(val); AmbienceLowPassCallback(); } , 0f, 22000f);
				CreateSlider(AmbienceLowPass, false);
				
				AmbienceEnableOcclusion = new JSONStorableBool("Enable occlusion", false);
				AmbienceEnableOcclusion.setCallbackFunction += (val) => { onEnableOcclusion(); };
				CreateToggle(AmbienceEnableOcclusion, false);
				
				OcclusionLowestFrequency = new JSONStorableFloat("Occlusion Lowest Frequency", 850f, 0f, 22000f);
				tmpSlider = CreateSlider(OcclusionLowestFrequency);
				OcclusionUIElements[0] = (UIDynamic)tmpSlider;
				
				OcclusionHighestFrequency = new JSONStorableFloat("Occlusion Highest Frequency", 18000f, 0f, 22000f);
				tmpSlider = CreateSlider(OcclusionHighestFrequency);
				OcclusionUIElements[1] = (UIDynamic)tmpSlider;
				
				OcclusionVolumeAttenuation = new JSONStorableFloat("Occlusion Volume Attenuation (%)", 20f, 0f, 50f);
				tmpSlider = CreateSlider(OcclusionVolumeAttenuation);
				OcclusionUIElements[2] = (UIDynamic)tmpSlider;
				
				OcclusionInterpolationSpeed = new JSONStorableFloat("Occlusion Interpolation Speed", 10f, 5f, 10f);
				tmpSlider = CreateSlider(OcclusionInterpolationSpeed);
				OcclusionUIElements[3] = (UIDynamic)tmpSlider;

				SoundWideness = new JSONStorableFloat("Sound Wideness", 1f, 0f, 3f);
				tmpSlider = CreateSlider(SoundWideness);
				OcclusionUIElements[4] = (UIDynamic)tmpSlider;
				
				ListenerWideness = new JSONStorableFloat("Listener Wideness", 0.05f, 0f, 0.5f);
				tmpSlider = CreateSlider(ListenerWideness);
				OcclusionUIElements[5] = (UIDynamic)tmpSlider;
				
				
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Help\nVAMAtmosphere Ambience</b></size></color>\n\n" + 
					"<size=30><color=#333>" +
					"<b>IMPORTANT:</b> If you switch asset in the Asset tab of the CUA, don't forget to reload the script to enable customization.\n\n" +
					"<b>Ambience max distance:</b> the maximum distance at which the ambience can be heard. You can check <b>Show bounds</b> below to see a gizmo previewing the distance.\n\n" +
					"<b>Ambience volume:</b> the volume of the ambience.\n\n" +
					"<b>Ambience pitch:</b> the pitch of the ambience. A negative value will play the ambience in reverse.\n\n" +
					"<b>Ambience spatialization:</b> if the sound is coming from its position in the world (3D / 5.1) or if it can be heard from anywhere (2D / Stereo).\n\n" +
					"<b>Ambience reverb:</b> how much the ambience sends its signal to reverb zones around it. You would have to use another CUA with the reverb zone or else this settings will have no apparent effect.\n\n" +
					"<b>Ambience high pass:</b> how much the ambience cuts the low frequencies. At 10 you can hear all the frequencies. At 22000 you will hear nothing. It is useful to remove the bass from the sound. <b>Ignored when using the realtime occlusion</b>\n\n" +
					"<b>Ambience low pass:</b> how much the ambience cuts the high frequencies. At 22000 you can hear all the frequencies. At 0 you will hear nothing. It is useful to simulate a <i>muffled effect</i>.\n\n" +
					"<b>Enable Occlusion:</b> Enables the occlusion system for this ambience.\n\n" +
					"<b>Occlusion Lowest Frequency:</b> How much the sound will be muffled. Lower values will make only the low frequencies (basses) hearable. <b>It should be lower than Ambience low pass and also lower than Occlusion Highest Frequency.</b>\n\n" +
					"<b>Occlusion Highest Frequency:</b> How much the sound will be muffled by default. Lower values will make less high frequencies (treble) hearable.\n\n" +
					"<b>Occlusion Volume Attenuation:</b> How much the occlusion will affect the volume of the sound.\n\n" +
					"<b>Occlusion Interpolation Speed:</b> How fast the occlusion will update. Values closer to 10 will make the change fast. Values closer to 5 slower.\n\n" +
					"<b>Sound Wideness:</b> How wide or large is the sound. Low values = a really directional sound, like a voice, or a phone ringing. High = a really big sound like loud music playing.\n\n" +
					"<b>Listener Wideness:</b> How wide or large is the player \"head\". Same as sound wideness but on the receiving hand.\n\n" +
					"<b>Show Bounds:</b> is a visual gizmo to help you tweak the maximum distance of the ambience.\n\n" +
					"<b>Show occlusion debug:</b> shows the raycasting debug for the occlusion.\n" +
					"</color></size>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 1000.0f;
				
				ShowDebug = new JSONStorableBool("Show bounds", false, ShowDebugCallback){ isStorable = false, isRestorable=false };
				CreateToggle(ShowDebug, true);
				
				ShowOcclusionDebug = new JSONStorableBool("Show occlusion debug", false){ isStorable = false, isRestorable=false };
				ShowOcclusionDebug.setCallbackFunction += (val) => { onUpdateOcclusionDebugDisplay(val); };
				CreateToggle(ShowOcclusionDebug, true);
				
				
				RegisterStringChooser(Ambience);
				RegisterFloat(AmbienceMaxDist);
				RegisterFloat(AmbienceVolume);
				RegisterFloat(AmbiencePitch);
				RegisterFloat(AmbienceSpatialization);
				RegisterFloat(AmbienceReverb);
				RegisterFloat(AmbienceHighPass);
				RegisterFloat(AmbienceLowPass);
				
				RegisterBool(AmbienceEnableOcclusion);
				RegisterFloat(OcclusionLowestFrequency);
				RegisterFloat(OcclusionHighestFrequency);
				RegisterFloat(OcclusionVolumeAttenuation);
				RegisterFloat(OcclusionInterpolationSpeed);
				RegisterFloat(SoundWideness);
				RegisterFloat(ListenerWideness);
            }
            catch(Exception e)
            {
                SuperController.LogError("VAMAAmbience - Exception caught: " + e);
            }
		}
		
		void Start()
        {

        }
		
		void Update() {
			// If we haven't made over 1min of init tests, we aren't initialized and the particle system (CUA) is still compatible, try to initialize
			if( initLoops < maxInitLoops && cuaInitialized == false ) {
				initLoops++;
				InitCUA();
			}
			
			if( initLoops == maxInitLoops ) {
				logError("VAMAAmbience : CUA could not be initialized properly after 3600 tries. Please select a compatible CUA and reload the script.");
				initLoops++; // To avoid spamming the log
			}
			
			if ( cuaInitialized == true && AmbienceEnableOcclusion.val == true )
			{
				TraceOcclusion(AmbienceRoot, CLTransform);
			}
		}


		// **************************
		// Functions
		// **************************
				
		private bool InitCUA() {
			if( cuaInitialized == true ) return true;
			
			try
			{
				CUARoot = getCUARoot( containingAtom.reParentObject );
				if( CUARoot == null ) {
					return false;
				}
				
				AmbienceRoot = getObjectNamed( CUARoot, "ambience_audio");
				if( AmbienceRoot == null ) {
					return false;
				}
							
				AsAmbience = AmbienceRoot.gameObject.GetComponent<AudioSource>();
				if( AsAmbience != null ) {
					AlpfAmbience = AmbienceRoot.gameObject.GetComponent<AudioLowPassFilter>();
					AhpfAmbience = AmbienceRoot.gameObject.GetComponent<AudioHighPassFilter>();
					
					Transform TrDebug = getObjectNamed( CUARoot, "debug_radius_max");
					GoADebugMaxDist = TrDebug.gameObject;
					CLTransform = SuperController.singleton.centerCameraTarget.targetCamera.transform;

					InitDebug();
					cuaInitialized = true;

					// Loading the bundle
					Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + "/vamatmosphere_ambiences.audiobundle", callback = OnAudioBundleLoaded};
					AssetLoader.QueueLoadAssetBundleFromFile(request);
				}
			}
			catch(Exception e)
			{
				SuperController.LogError("VAMAAmbience - Exception caught: " + e);
			}
			
			return false;

		}
		
		protected void InitPaths() {
			// Creating assets paths list
			ambiencesPaths = new List<string>();
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_bar_south_africa.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_bar_us.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_birds_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_city_birds_dogs_cars.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_city_night.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_city_urban.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_ethereal_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_forest_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_forest_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_night_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_night_strange_jungle.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_ocean_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_party_germany.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rain_heavy_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rain_heavy_city_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rain_heavy_city_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rain_light_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rain_thunder_city_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_river_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rumble_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rumble_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scary_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scary_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scary_03.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scifi_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scifi_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scifi_03.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_city_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_city_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_shower_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_shower_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_shower_03.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_shower_04.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_underwater_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_underwater_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_vent_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_wind_01.ogg");
		}
		
		private void UpdateAmbienceSettings() {
			AmbienceCallback( Ambience.val );
			
			AmbienceMaxDistCallback(AmbienceMaxDist.val);
			AmbienceVolumeCallback(AmbienceVolume.val);
			AmbiencePitchCallback(AmbiencePitch.val);
			AmbienceSpatializationCallback(AmbienceSpatialization.val);
			AmbienceReverbCallback(AmbienceReverb.val);
			AmbienceHighPassCallback();
			AmbienceLowPassCallback();
			
			ShowDebugCallback( ShowDebug.val );
			onEnableOcclusion();
		}
		
		protected void UpdateDebugDisplay() {
			if ( GoADebugMaxDist != null ) {
				GoADebugMaxDist.transform.localScale = new Vector3( AmbienceMaxDist.val, AmbienceMaxDist.val, AmbienceMaxDist.val );
			}
		}
		
		private void InitDebug()
		{
			_debugLinesList = new LineRenderer[maxDebugLines];
			for (int i = 0; i < maxDebugLines; i++)
			{
				GameObject lineObj = new GameObject("debugLine" + i);
				LineRenderer lineRenderer = lineObj.AddComponent<LineRenderer>();
				lineRenderer.material = new Material(Shader.Find("Hidden/Internal-Colored"));
				lineRenderer.startColor = Color.green;
				lineRenderer.endColor = Color.green;
				lineRenderer.startWidth = 0.0001f;
				lineRenderer.endWidth = 0.0001f;
				lineRenderer.positionCount = 2;
				lineRenderer.gameObject.SetActive(false);
				_debugLinesList[i] = lineRenderer;
			}
		}
		
		private void showDebugLine(Vector3 start, Vector3 end, Color newColor)
		{
			if (ShowOcclusionDebug.val == false || _debugLinesList == null) return;
			
			// What iz diz you ask me ? I'm compensating a bit the vertical offset because of the player camera
			// Otherwise you don't see the whole debug since it's casted straight to the center of the view
			end.y -= 0.01f;
			
			_debugLinesList[debugLineCount].gameObject.SetActive(true);
			_debugLinesList[debugLineCount].SetPosition(0, start);
			_debugLinesList[debugLineCount].SetPosition(1, end);
			_debugLinesList[debugLineCount].startColor = newColor;
			_debugLinesList[debugLineCount].endColor = newColor;
			debugLineCount++;
			if (debugLineCount >= maxDebugLines) debugLineCount = 0;
		}
		
		private void TraceOcclusion(Transform SourceTR, Transform ListenerTR)
		{
			// To prevent crash because of the CUA
			if (AmbienceRoot == null)
			{
				ShowOcclusionDebug.val = false;
				return;
			}

			// Computation
			raycastTotal = 0;
			raycastHitCount = 0;
			Vector3 sourceLoc = SourceTR.position;
			Vector3 listenerLoc = ListenerTR.position;
			
			Vector3 sourceLeft = CalculatePoint(sourceLoc, listenerLoc, SoundWideness.val, true);
			Vector3 sourceRight = CalculatePoint(sourceLoc, listenerLoc, SoundWideness.val, false);

			Vector3 listenerLeft = CalculatePoint(listenerLoc, sourceLoc, ListenerWideness.val, true);
			Vector3 listenerRight = CalculatePoint(listenerLoc, sourceLoc, ListenerWideness.val, false);

			CastLine(sourceLeft, listenerLeft);
			CastLine(sourceLeft, listenerLoc);
			CastLine(sourceLeft, listenerRight);

			CastLine(sourceLoc, listenerLeft);
			CastLine(sourceLoc, listenerLoc);
			CastLine(sourceLoc, listenerRight);

			CastLine(sourceRight, listenerLeft);
			CastLine(sourceRight, listenerLoc);
			CastLine(sourceRight, listenerRight);
			
			ComputeOcclusionFactorFast();
		}

		private void ComputeOcclusionFactorFast()
		{
			// Default values
			newCutoff = AmbienceLowPass.val;
			newVolumeAtt = AmbienceVolume.val;
			// Computing changes if there is at least one hit
			if ( raycastHitCount > 0 )
			{
				// Cutoff
				// I'm dividing the amoumnt of cutoff we want but the total raycast done
				// Since if we'd have all the cast occluded, it would be a complete occlusion
				// then multiplying by the number of hits 
				float occHighestFreq = Mathf.Min(AmbienceLowPass.val, OcclusionHighestFrequency.val); // To prevent someone putting a highest frequency in the occlusion settings than the LPF in the basic settings
				float occLowestFreq = Mathf.Min(AmbienceLowPass.val, OcclusionLowestFrequency.val); // To prevent someone putting a highest frequency in the occlusion settings than the LPF in the basic settings
				float occlusionStep = ((occHighestFreq - occLowestFreq) / raycastTotal);
				newCutoff = occHighestFreq - (occlusionStep * raycastHitCount);

				// Volume attenuation
				newVolumeAtt = AmbienceVolume.val - ((AmbienceVolume.val * OcclusionVolumeAttenuation.val / 100) / raycastTotal) * raycastHitCount;
			}

			// ******************
			// Setting the values
			// ******************
			// cutoff
			if( AlpfAmbience != null ) AlpfAmbience.cutoffFrequency = Mathf.Lerp(AlpfAmbience.cutoffFrequency, newCutoff, OcclusionInterpolationSpeed.val * Time.deltaTime);
			// volume
			AsAmbience.volume = Mathf.Lerp( AsAmbience.volume, newVolumeAtt, OcclusionInterpolationSpeed.val * Time.deltaTime);
		}
		
		private Vector3 CalculatePoint(Vector3 a, Vector3 b, float m, bool posOrneg, float vertOffset=0f)
		{
			float x;
			float z;
			float n = Vector3.Distance(new Vector3(a.x, 0f, a.z), new Vector3(b.x, 0f, b.z));
			float mn = (m / n);
			if (posOrneg)
			{
				x = a.x + (mn * (a.z - b.z));
				z = a.z - (mn * (a.x - b.x));
			}
			else
			{
				x = a.x - (mn * (a.z - b.z));
				z = a.z + (mn * (a.x - b.x));
			}
			return new Vector3(x, a.y + vertOffset, z);
		}
		
		private void CastLine(Vector3 Start, Vector3 End)
		{
			RaycastHit hit;
			RaycastHit hitBack;
			// Seems like the layer "0" is on what mostly anything that should hit is on (geo, meshes, CUAs...)
			// Based on Unity's doc, our layer mask is 1 << layerID, so 1 << 0 = 1
			Physics.Linecast(Start, End, out hit, 1);
			raycastTotal++;
			// Do we have a hit ?
			if ( hit.collider )
			{
				raycastHitCount++;
				showDebugLine(Start, End, Color.red);
			}
			else
				showDebugLine(Start, End, Color.green);
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		private void OnAudioBundleLoaded(Request aRequest) {
			try
            {
				InitPaths();
				
				audioBundleRequest = aRequest;
				
				ambienceChoices = new List<string>();
				foreach( string path in ambiencesPaths ) {
					string[] aPath = path.Split('/');
					string[] aFile = aPath.Last().Split('.');
					string fileName = aFile[aFile.Length - 2].ToString();
					ambienceChoices.Add( fileName );
				}
				
				Ambience.choices = ambienceChoices;
				
				UpdateAmbienceSettings();
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAAmbience - Exception caught: " + e);
            }
		}
		
		protected void AmbienceCallback(string choice) {
			if( CUARoot && AsAmbience ) {
				// I need to prevent a callback if the lists are not ready
				if( ambiencesPaths != null && ambienceChoices != null && ambiencesPaths.Count > 0 && ambienceChoices.Count > 0 && choice != "" ) {
					try{
						AudioClip ac = audioBundleRequest.assetBundle.LoadAsset<AudioClip>(ambiencesPaths[ ambienceChoices.IndexOf(choice) ]);
						AsAmbience.clip = ac;
						
						// Defining the start point of the sound to avoid chorus / echo effect (and also randomness!)
						float startTime = UnityEngine.Random.Range(0, Mathf.Round( ac.length / 2 ));
						AsAmbience.time = startTime;
						
						AsAmbience.volume = 0f;
						StartCoroutine(FadeInAmbience(AsAmbience,2f,AmbienceVolume.val));
						
						AsAmbience.Play();
					}
					catch(Exception e)
					{
						SuperController.LogError("VAMAAmbience - Exception caught: " + e);
					}
				}
			}
		}
		
		protected void AmbienceMaxDistCallback(float size) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.maxDistance = size;
				
				UpdateDebugDisplay();
			}
		}
		
		protected void AmbienceVolumeCallback(float volume) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.volume = volume;
			}
		}


		protected void AmbiencePitchCallback(float pitch) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.pitch = pitch;
			}
		}
		
		protected void AmbienceSpatializationCallback(float val) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.spatialBlend = val;
			}
		}
		
		protected void AmbienceReverbCallback(float val) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.reverbZoneMix = val;
			}
		}
		
		protected void AmbienceHighPassCallback() {
			if( CUARoot && AhpfAmbience ) {
				AhpfAmbience.cutoffFrequency = AmbienceHighPass.val;
			}
		}
		
		protected void AmbienceLowPassCallback() {
			if( CUARoot && AlpfAmbience ) {
				AlpfAmbience.cutoffFrequency = AmbienceLowPass.val;
			}
		}
		
		protected void ShowDebugCallback(bool enabled) {
			if( CUARoot ) {				
				if( enabled == true ) {
					GoADebugMaxDist.SetActive(true);			
					
					UpdateDebugDisplay();
				} else {
					GoADebugMaxDist.SetActive(false);
					
					UpdateDebugDisplay();
				}
			}
		}
		
		protected void onUpdateOcclusionDebugDisplay( bool activeState )
		{
			// No debug lines if we don't have the occlusion enabled
			if (activeState == true && AmbienceEnableOcclusion.val == false) return;
			// Or changing their state
			if (_debugLinesList != null)
			{
				for (int i = 0; i < maxDebugLines; i++)
				{
					_debugLinesList[i].gameObject.SetActive(activeState);
				}
			}
		}

		protected void onEnableOcclusion()
		{
			Vector3 occUiVecVal = new Vector3(1,1,1);
			bool occUiLayoutVal = false;
			if (AmbienceEnableOcclusion.val == false)
			{
				occUiVecVal = new Vector3(0,0,0);
				occUiLayoutVal = true;
			}
			
			// Reset when disabled
			if (AmbienceEnableOcclusion.val == false)
			{
				AsAmbience.volume = AmbienceVolume.val;
				AlpfAmbience.cutoffFrequency = AmbienceLowPass.val;
				ShowOcclusionDebug.val = false;
			}
			
			foreach (UIDynamic uidElem in OcclusionUIElements)
			{
				uidElem.GetComponent<LayoutElement>().transform.localScale = occUiVecVal;
				uidElem.GetComponent<LayoutElement>().ignoreLayout = occUiLayoutVal;
			}
		}
			
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
        // Recursive function searching for a transform
        private Transform getCUARoot( Transform parent ) {
            foreach( Transform child in parent )
            {
				if( child.name.Contains("VAMAAmbience") ) {
					return child;
				} else {
					Transform childSearch = getCUARoot(child);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		private Transform getObjectNamed( Transform parent, string objName ) {
            foreach( Transform child in parent )
            {
				if( child.name == objName ) {
					return child;
				} else {
					Transform childSearch = getObjectNamed(child,objName);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		// **************************
		// COROUTINES
		// **************************
		public static IEnumerator FadeInAmbience(AudioSource audioSource, float fadeDuration, float targetVolume)
		{
			// No audio source ? stop this immediately
			if (audioSource == null) yield break;
			
			float currentTime = 0f;
			float start = 0f;

			while (currentTime < fadeDuration)
			{
				// No audio source after we entered the loop ? stop this immediately
				if (audioSource == null) yield break;
				
				currentTime += Time.deltaTime;
				audioSource.volume = Mathf.Lerp(start, targetVolume, currentTime / fadeDuration);
				yield return null;
			}
		}
		
		public static IEnumerator FadeOutAndStopAmbience(AudioSource audioSource, float fadeDuration)
		{
			// No audio source ? stop this immediately
			if (audioSource == null) yield break;
			
			float currentTime = 0f;
			float start = audioSource.volume;

			while (currentTime < fadeDuration)
			{
				// No audio source after we entered the loop ? stop this immediately
				if (audioSource == null) yield break;
				
				currentTime += Time.deltaTime;
				audioSource.volume = Mathf.Lerp(start, 0f, currentTime / fadeDuration);
				yield return null;
			}
			
			audioSource.Stop();
		}
		
		// **************************
		// Time to cleanup !
		// **************************
		void CleanupScene()
		{
			ShowOcclusionDebug.val = false;

			// Cleaning debug lines
			if (_debugLinesList != null)
			{
				for (int i = 0; i < maxDebugLines; i++)
				{
					Destroy(_debugLinesList[i].gameObject);
				}
			}
		}
		
		void OnEnable() {
			// Fading in
			AsAmbience.volume = 0f;
			AsAmbience.Play();
			StartCoroutine(FadeInAmbience(AsAmbience,2f,AmbienceVolume.val));
			
			// Disabling debug
			ShowDebugCallback( ShowDebug.val );
		}
		
		void OnDisable() {
			// Fading out
			if (AsAmbience != null)
			{
				AsAmbience.volume = 0f;
				AsAmbience.Stop();
			}

			ShowOcclusionDebug.val = false;
			
			// Disabling debug
			ShowDebugCallback( false );
		}
		
		void OnDestroy() {			
			// Disabling debug
			ShowDebugCallback( false );
			
			CleanupScene();
			
			// Cleaning modified native VAM assets
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + "/vamatmosphere_ambiences.audiobundle");
		}

		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		

	}
}
