using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMAtmosphere - Occlusion v1.0
//
// An occlusion system for AudioSource and Characters
// Based on:
// - Deadspace devlog: https://www.youtube.com/watch?v=yMQOkpZO5eM
// - Quantum break devlog and GDC: https://www.dailymotion.com/video/x2wgj2f - https://www.gdcvault.com/play/1022402/Using-Umbra-Spatial-Data-for
// - Scott game sounds: https://www.youtube.com/watch?v=wTOHc803_ys

namespace VAMAOcclusionPlugin
{
    public class VAMAOcclusion : MVRScript
    {	
        public static string PLUGIN_PATH;

		private GameObject GoADebugMaxDist;
		
		private Transform CASTransform; // Current Audio Source Transform
		private Transform CLTransform; // Current Listener Transform
		private AudioSource CurrentAudioSource;
		private AudioSourceControl CurrentAudioSourceControl;
		
		private AudioLowPassFilter CASLowPassFilter;
		
		// Misc
		private float internalTimer = 0f;
		private bool SoundSourceInitialized = false;
		private bool DebugEnabled = false;
		private float stateUpdateTime = 0.5f;
		private float nextStateUpdate = 0f;
		private bool shitHappened = false;
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		
		// Bench
		private System.Diagnostics.Stopwatch benchStopWatch = new System.Diagnostics.Stopwatch();
		private double[] benchAvgUpdateTimes = new double[80];
		private double benchAvgUpdateTime = 0;
		
		// Raycasting variables
		private LineRenderer[] _debugLinesList;
		private int debugLineCount = 0;
		private static int maxDebugLines = 13;
		private int raycastTotal = 0;
		private int raycastHitCount = 0;
		private float hitThicknessTotal = 0; // the total width of the trace + backtrace to the source
		private float sourceDistance = 0f;
		private float newCutoff = 0f;
		private float newVolumeAtt = 0f;
		private float newSpatialBlend = 0f;
		private float hitContribution = 0f;
		private float thicknessContribution = 0f;

		// JSONStorables
		public JSONStorableFloat AudioSourceVolume;
		public JSONStorableFloat ProximityBleed;
		public JSONStorableFloat ProximityBleedDistance;
		public JSONStorableStringChooser OcclusionType;
		public JSONStorableFloat OcclusionThicknessContribution;
		public JSONStorableFloat OcclusionThickness;
		public JSONStorableFloat OcclusionLowestFrequency;
		public JSONStorableFloat OcclusionHighestFrequency;
		public JSONStorableFloat OcclusionVolumeAttenuation;
		public JSONStorableFloat OcclusionInterpolationSpeed;
		public JSONStorableFloat SoundWideness;
		public JSONStorableFloat ListenerWideness;
		public JSONStorableBool ShowDebug;
		public JSONStorableBool ShowStatistics;
		public JSONStorableBool LockDebugLines;
		private JSONStorableString stateInfos;

		protected List<string> OcclusionTypeChoices;
		
		// External plugins
		private JSONStorable VAMMoanPlugin;
		
		// UI Elements
		private UIDynamic HelpUIElement;
		private UIDynamic[] fastTypeUIElements;
		private UIDynamic[] complexTypeUIElements;

		public override void Init()
        {			 
            try
            {
                if ( containingAtom.type != "AudioSource" && containingAtom.type != "Person" && containingAtom.type != "AptSpeaker" && containingAtom.type != "CustomUnityAsset" )
                {
                    SuperController.LogError("Please add VAMAOcclusion on an AudioSource, a character, an AptSpeaker or a CustomUnityAsset with a VAMAAtmosphere plugin.");
                    return;
                }

				PLUGIN_PATH = GetPluginPath(this);
				
				// Base variables for the UI
				fastTypeUIElements = new UIDynamic[1];
				complexTypeUIElements = new UIDynamic[2];
				UIDynamicSlider tmpSlider;
				
				// Creating the JSONStorables
				AudioSourceVolume = new JSONStorableFloat("Volume", 1f, 0f, 1f, false);
				AudioSourceVolume.setCallbackFunction += (val) => { UpdateSettings(); };
				CreateSlider(AudioSourceVolume);
				
				OcclusionTypeChoices = new List<string>();
				OcclusionTypeChoices.Add("Fast");
				OcclusionTypeChoices.Add("Complex");
				
				OcclusionType = new JSONStorableStringChooser("Occlusion Type", OcclusionTypeChoices, "Fast", "Occlusion Type");
				OcclusionType.setCallbackFunction += (val) => { onChangeOcclusionType(); };
				CreateScrollablePopup(OcclusionType);
				
				OcclusionThicknessContribution = new JSONStorableFloat("Occlusion Thickness Contribution (%)", 40f, 20f, 70f);
				OcclusionThicknessContribution.setCallbackFunction += (val) => { onChangeOcclusionThicknessContribution(); };
				tmpSlider = CreateSlider(OcclusionThicknessContribution);
				complexTypeUIElements[0] = (UIDynamic)tmpSlider;
				
				OcclusionThickness = new JSONStorableFloat("Occlusion Thickness", 4f, 3f, 10f);
				tmpSlider = CreateSlider(OcclusionThickness);
				complexTypeUIElements[1] = (UIDynamic)tmpSlider;
				
				OcclusionLowestFrequency = new JSONStorableFloat("Occlusion Lowest Frequency", 1500f, 0f, 8000f);
				tmpSlider = CreateSlider(OcclusionLowestFrequency);
				fastTypeUIElements[0] = (UIDynamic)tmpSlider;
				
				OcclusionHighestFrequency = new JSONStorableFloat("Occlusion Highest Frequency", 18000f, 8000f, 22000f);
				CreateSlider(OcclusionHighestFrequency);
				
				OcclusionVolumeAttenuation = new JSONStorableFloat("Occlusion Volume Attenuation (%)", 20f, 0f, 50f);
				CreateSlider(OcclusionVolumeAttenuation);
				
				OcclusionInterpolationSpeed = new JSONStorableFloat("Occlusion Interpolation Speed", 10f, 5f, 10f);
				CreateSlider(OcclusionInterpolationSpeed);
				
				ProximityBleed = new JSONStorableFloat("Proximity Bleed (%)", 0f, 0f, 50f, false);
				CreateSlider(ProximityBleed);
				
				ProximityBleedDistance = new JSONStorableFloat("Proximity Bleed Distance", 2f, 2f, 10f, false);
				CreateSlider(ProximityBleedDistance);
				
				SoundWideness = new JSONStorableFloat("Sound Wideness", 1f, 0f, 3f);
				CreateSlider(SoundWideness);
				
				ListenerWideness = new JSONStorableFloat("Listener Wideness", 0.05f, 0f, 0.5f);
				CreateSlider(ListenerWideness);
				
				ShowDebug = new JSONStorableBool("Show debug", false);
				ShowDebug.setCallbackFunction += (val) => { onUpdateDebugDisplay(val); };
				CreateToggle(ShowDebug, false);
				
				LockDebugLines = new JSONStorableBool("Lock debug lines", false);
				CreateToggle(LockDebugLines, false);
				
				
				// UI RIGHT
				ShowStatistics = new JSONStorableBool("Show statistics", false);
				ShowStatistics.setCallbackFunction += (val) => { onShowStatistics(); };
				CreateToggle(ShowStatistics, true);
		
				
				stateInfos = new JSONStorableString("Stateinfos", "");
				HelpUIElement = CreateTextField(stateInfos, true);
				HelpUIElement.height = 215.0f;
			
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Help - Occlusion</b></size></color>\n\n" + 
					"<size=30><color=#333>" +
					"This plugin allows realtime occlusion of sounds, it can be used on AudioSource or Characters.\n\n" +
					"<b>Volume:</b> Controls the AudioSource volume.\n\n" +
					"<b>Occlusion Type:</b> How the occlusion is computed. <b>Fast</b> only relies on multiple hit from the source. <b>Complex</b> relies on hits from the source and a backtrace to compute the thickness of objects between the source and you. Using <b>Complex</b> is more detailed and accurate but is more intensive (Cost is double as Fast). Use it for sounds you think are important.\n\n" +
					"<b>Occlusion Thickness Contribution:</b> How much the thickness of the materials between you and the source contributes to occluding the sound. High values will make the thickness more important than the simple fact that you have an object in front of you. <b>Only on Complex occlusion type.</b>\n\n" +
					"<b>Occlusion Thickness:</b> The thickness value in meters when the sound reaches maximum occlusion. Lower value will make the sound more occluded with thin walls and vice versa. <b>Only on Complex occlusion type.</b>\n\n" +
					"<b>Occlusion Lowest Frequency:</b> How much the sound will be muffled. Lower values will make only the low frequencies (basses) hearable. <b>Only on Fast occlusion type.</b>\n\n" +
					"<b>Occlusion Highest Frequency:</b> How much the sound will be muffled by default. Lower values will make less high frequencies (treble) hearable.\n\n" +
					"<b>Occlusion Volume Attenuation:</b> How much the occlusion will affect the volume of the sound.\n\n" +
					"<b>Occlusion Interpolation Speed:</b> How fast the occlusion will update. Values closer to 10 will make the change fast. Values closer to 5 slower.\n\n" +
					"<b>Proximity Bleed:</b> Simulate the \"bleed\" and reflection of sound in a room, the higher the value, the more the sound will appear wide and coming from everywhere.\n\n" +
					"<b>Proximity Bleed Distance:</b> The distance from the source at which the \"bleed\" starts to occur. The sound will become wider the closer you are to it. You would only use this for wide sound (music for instance).\n\n" +
					"<b>Sound Wideness:</b> How wide or large is the sound. Low values = a really directional sound, like a voice, or a phone ringing. High = a really big sound like loud music playing.\n\n" +
					"<b>Listener Wideness:</b> How wide or large is the player \"head\". Same as sound wideness but on the receiving hand.\n\n" +
					"<b>Show debug:</b> Show the line trace from the source to the player.\n\n" +
					"<b>Lock debug lines:</b> Locks the debug line to check their positions.\n" +
					"</color></size>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 1150.0f;
				
				// Registering JSONStorables
				RegisterFloat(AudioSourceVolume);
				RegisterStringChooser(OcclusionType);
				RegisterFloat(OcclusionThicknessContribution);
				RegisterFloat(OcclusionThickness);
				RegisterFloat(OcclusionLowestFrequency);
				RegisterFloat(OcclusionHighestFrequency);
				RegisterFloat(OcclusionVolumeAttenuation);
				RegisterFloat(OcclusionInterpolationSpeed);
				RegisterFloat(ProximityBleed);
				RegisterFloat(ProximityBleedDistance);
				RegisterFloat(SoundWideness);
				RegisterFloat(ListenerWideness);

            }
            catch(Exception e)
            {
                SuperController.LogError("VAMAOcclusion - Exception caught: " + e);
            }
		}
		
		void Start()
        {
	        
        }
		
		void Update() {
			internalTimer += Time.deltaTime;
			
			// If the editor is not initialized, we're gonna try to do it
			// I'm putting 3600 tries ( at 60fps it is 1min )
			if (initLoops < maxInitLoops && SoundSourceInitialized == false)
			{
				initLoops++;
				InitSoundSource();
			}
			
			// Triggering an error if we have reached our limit of tries
			if( initLoops == maxInitLoops && SoundSourceInitialized == false ) {
				initLoops++; // To avoid spamming the log with the error
				SuperController.LogError("VAMAOcclusion - Error:\n The occlusion could not initialize properly. Maybe you haven't put the occlusion plugin on the right atom?");
			}

			if ( SoundSourceInitialized == true)
			{
				TraceOcclusion(CASTransform, CLTransform);
				UpdateStateInfos();
			}

			if (shitHappened == true)
			{
				SuperController.LogError("VAMAOcclusion - Error:\nThe Low Pass Filter is gone!!!  WHY? :'(\nThis is a well know bug when using the reload button. (and it seems like this is some kind of Unity mystery sadly)\nPlease reload it once more to fix this or try removing and adding the plugin again.");
				shitHappened = false;
			}
		}


		// **************************
		// Functions
		// **************************
				
		private bool InitSoundSource()
		{
			try
			{
				if (containingAtom.type == "Person")
				{
					CurrentAudioSourceControl = containingAtom.GetStorableByID("HeadAudioSource") as AudioSourceControl;
					CurrentAudioSource = CurrentAudioSourceControl.audioSource;
					CASTransform = CurrentAudioSourceControl.audioSource.transform;
					VAMMoanPlugin = containingAtom.GetStorableIDs().Select(containingAtom.GetStorableByID).FirstOrDefault(s => s.IsFloatJSONParam("VAMM Global Occlusion LPF"));
				}
				else if(containingAtom.type == "AptSpeaker")
				{
					CurrentAudioSourceControl = containingAtom.GetStorableByID("AptSpeaker_Import") as AudioSourceControl;
					CurrentAudioSource = CurrentAudioSourceControl.audioSource;
					CASTransform = CurrentAudioSourceControl.audioSource.transform;
				}
				else
				{
					CurrentAudioSourceControl = containingAtom.GetStorableByID("AudioSource") as AudioSourceControl;
					CurrentAudioSource = CurrentAudioSourceControl.audioSource;
					CASTransform = CurrentAudioSourceControl.audioSource.transform;
				}
				


				CASLowPassFilter = CurrentAudioSource.gameObject.GetComponent<AudioLowPassFilter>();
				if (CASLowPassFilter == null)
				{
					CASLowPassFilter = CurrentAudioSource.gameObject.AddComponent<AudioLowPassFilter>();
				}
				CASLowPassFilter.cutoffFrequency = 22000.0f; // Default, no cutoff
				CLTransform = SuperController.singleton.centerCameraTarget.targetCamera.transform;

				if (CASLowPassFilter != null && CurrentAudioSource != null && CASTransform != null)
				{
					SoundSourceInitialized = true;
					InitDebug();
					onShowStatistics();
					onChangeOcclusionType();

					return true;
				}
			}
			catch(Exception e)
			{
				SuperController.LogError("VAMAOcclusion - Exception caught: " + e);
			}

			return false;

		}

		private void InitDebug()
		{
			_debugLinesList = new LineRenderer[maxDebugLines];
			for (int i = 0; i < maxDebugLines; i++)
			{
				GameObject lineObj = new GameObject("debugLine" + i);
				LineRenderer lineRenderer = lineObj.AddComponent<LineRenderer>();
				lineRenderer.material = new Material(Shader.Find("Hidden/Internal-Colored"));
				lineRenderer.startColor = Color.green;
				lineRenderer.endColor = Color.green;
				lineRenderer.startWidth = 0.0001f;
				lineRenderer.endWidth = 0.0001f;
				lineRenderer.positionCount = 2;
				lineRenderer.gameObject.SetActive(false);
				_debugLinesList[i] = lineRenderer;
			}
		}

		private void UpdateSettings()
		{
			if( CurrentAudioSourceControl != null )
				CurrentAudioSourceControl.SetFloatParamValue("volume",AudioSourceVolume.val);
		}

		private void showDebugLine(Vector3 start, Vector3 end, Color newColor)
		{
			if (DebugEnabled == false || LockDebugLines.val == true) return;
			
			// What iz diz you ask me ? I'm compensating a bit the vertical offset because of the player camera
			// Otherwise you don't see the whole debug since it's casted straight to the center of the view
			end.y -= 0.01f;
			
			_debugLinesList[debugLineCount].gameObject.SetActive(true);
			_debugLinesList[debugLineCount].SetPosition(0, start);
			_debugLinesList[debugLineCount].SetPosition(1, end);
			_debugLinesList[debugLineCount].startColor = newColor;
			_debugLinesList[debugLineCount].endColor = newColor;
			debugLineCount++;
			if (debugLineCount >= maxDebugLines) debugLineCount = 0;
		}

		private void TraceOcclusion(Transform SourceTR, Transform ListenerTR)
		{
			// Bench
			StartBenchMark();
			
			// Computation
			raycastTotal = 0;
			raycastHitCount = 0;
			hitThicknessTotal = 0;
			Vector3 sourceLoc = SourceTR.position;
			
			// Super hacky compensation for the speaker itself.
			// The audio source is based on the floor position of the speaker, so I'm just offseting a bit
			// to avoid having the linecast starting from the floor (which might produce hits that are not wanted)
			if (containingAtom.type == "AptSpeaker")
			{
				sourceLoc.x -= 0.2f;
				sourceLoc.y += 0.7f;
				sourceLoc.z -= 0.2f;
			}
			
			Vector3 listenerLoc = ListenerTR.position;
			sourceDistance = Vector3.Distance(SourceTR.position, ListenerTR.position);
			
			Vector3 sourceLeft = CalculatePoint(sourceLoc, listenerLoc, SoundWideness.val, true);
			Vector3 sourceRight = CalculatePoint(sourceLoc, listenerLoc, SoundWideness.val, false);
			
			Vector3 sourceTopLeft = CalculatePoint(sourceLoc, listenerLoc, SoundWideness.val * 0.5f, true, SoundWideness.val * 0.3f);
			Vector3 sourceBotRight = CalculatePoint(sourceLoc, listenerLoc, SoundWideness.val * 0.5f, false, -SoundWideness.val * 0.3f);

			Vector3 listenerLeft = CalculatePoint(listenerLoc, sourceLoc, ListenerWideness.val, true);
			Vector3 listenerRight = CalculatePoint(listenerLoc, sourceLoc, ListenerWideness.val, false);
			
			Vector3 listenerTopLeft = CalculatePoint(listenerLoc, sourceLoc, ListenerWideness.val * 0.5f, true, ListenerWideness.val * 0.3f);
			Vector3 listenerBotRight = CalculatePoint(listenerLoc, sourceLoc, ListenerWideness.val * 0.5f,false, -ListenerWideness.val * 0.3f);

			CastLine(sourceLeft, listenerLeft);
			CastLine(sourceLeft, listenerLoc);
			CastLine(sourceLeft, listenerRight);

			CastLine(sourceLoc, listenerLeft);
			CastLine(sourceLoc, listenerLoc);
			CastLine(sourceLoc, listenerRight);

			CastLine(sourceRight, listenerLeft);
			CastLine(sourceRight, listenerLoc);
			CastLine(sourceRight, listenerRight);
        
			CastLine(sourceTopLeft, listenerTopLeft);
			CastLine(sourceTopLeft, listenerBotRight);
			
			CastLine(sourceBotRight, listenerTopLeft);
			CastLine(sourceBotRight, listenerBotRight);

			if ( OcclusionType.val == "Fast" )
			{
				ComputeOcclusionFactorFast();
			}
			else
			{
				ComputeOcclusionFactorComplex();
			}
			// Bench stop and calculation
			StopBenchmark();

			if (CASLowPassFilter == null)
			{
				SoundSourceInitialized = false;
				shitHappened = true;
			}
		}

		private void ComputeOcclusionFactorFast()
		{
			// Default values
			newCutoff = 22000f;
			newVolumeAtt = AudioSourceVolume.val;
			newSpatialBlend = 1f;
			// Computing changes if there is at least one hit
			if ( raycastHitCount > 0 )
			{
				// Cutoff
				// I'm dividing the amoumnt of cutoff we want but the total raycast done
				// Since if we'd have all the cast occluded, it would be a complete occlusion
				// then multiplying by the number of hits 
				float occlusionStep = ((OcclusionHighestFrequency.val - OcclusionLowestFrequency.val) / raycastTotal);
				newCutoff = OcclusionHighestFrequency.val - (occlusionStep * raycastHitCount);

				// Volume attenuation
				newVolumeAtt = AudioSourceVolume.val - ((AudioSourceVolume.val * OcclusionVolumeAttenuation.val / 100) / raycastTotal) * raycastHitCount;

				// Bleed
				// The closer you are to the more bleed you have
				// The calculation is capped at a min distance of 1
				if (ProximityBleed.val > 0f && sourceDistance > 1f)
				{
					if (sourceDistance <= ProximityBleedDistance.val)
					{
						newSpatialBlend = (ProximityBleed.val / (ProximityBleedDistance.val - 1f) * (ProximityBleedDistance.val - sourceDistance));
						newSpatialBlend = 1 - (newSpatialBlend / 100);
					}
				}
			}

			// ******************
			// Setting the values
			// ******************
			// cutoff
			if( CASLowPassFilter != null ) CASLowPassFilter.cutoffFrequency = Mathf.Lerp(CASLowPassFilter.cutoffFrequency, newCutoff, OcclusionInterpolationSpeed.val * Time.deltaTime);
			if (VAMMoanPlugin != null)
			{
				float VAMMLPFVal = Mathf.Lerp(VAMMoanPlugin.GetFloatParamValue("VAMM Global Occlusion LPF"), newCutoff, OcclusionInterpolationSpeed.val * Time.deltaTime);				
				VAMMoanPlugin.SetFloatParamValue("VAMM Global Occlusion LPF", VAMMLPFVal);
			}
			if (CurrentAudioSourceControl != null)
			{
				// volume
				CurrentAudioSourceControl.SetFloatParamValue("volume", Mathf.Lerp(CurrentAudioSourceControl.GetFloatParamValue("volume"), newVolumeAtt, OcclusionInterpolationSpeed.val * Time.deltaTime));
				// spatialBlend
				CurrentAudioSourceControl.SetFloatParamValue("spatialBlend", Mathf.Lerp(CurrentAudioSourceControl.GetFloatParamValue("spatialBlend"), newSpatialBlend, OcclusionInterpolationSpeed.val * Time.deltaTime));
			}
			else if ( CurrentAudioSource != null )
			{
				// volume
				CurrentAudioSource.volume = Mathf.Lerp( CurrentAudioSource.volume, newVolumeAtt, OcclusionInterpolationSpeed.val * Time.deltaTime);
				// spatialBlend
				CurrentAudioSource.spatialBlend = Mathf.Lerp( CurrentAudioSource.spatialBlend, newSpatialBlend, OcclusionInterpolationSpeed.val * Time.deltaTime);
			}
		}
		
		private void ComputeOcclusionFactorComplex()
		{
			// Default values
			newCutoff = 22000f;
			newVolumeAtt = AudioSourceVolume.val;
			newSpatialBlend = 1f;
			// Computing changes if there is at least one hit
			if ( raycastHitCount > 0 )
			{
				// Computing the average thickness remapping to "percents"
				float thicknessVal = raycastHitCount > 0 ? Mathf.Clamp( Remap(hitThicknessTotal / raycastHitCount, 0, OcclusionThickness.val, 0, 1), 0, 1 ) : 0;
				
				// Cutoff
				// I'm dividing the amoumnt of cutoff we want but the total raycast done
				// Since if we'd have all the cast occluded, it would be a complete occlusion
				// then multiplying by the number of hits

				float occlusionStep = ( ( OcclusionHighestFrequency.val * hitContribution ) / raycastTotal );
				float hitOcclusion = occlusionStep * raycastHitCount;
				float thicknessOcclusion = ( OcclusionHighestFrequency.val * thicknessContribution ) * thicknessVal;
				newCutoff = OcclusionHighestFrequency.val - ( hitOcclusion + thicknessOcclusion );

				// Volume attenuation
				newVolumeAtt = AudioSourceVolume.val - ((AudioSourceVolume.val * OcclusionVolumeAttenuation.val / 100) / raycastTotal) * raycastHitCount;

				// Bleed
				// The closer you are to the more bleed you have
				// The calculation is capped at a min distance of 1
				if (ProximityBleed.val > 0f && sourceDistance > 1f)
				{
					if (sourceDistance <= ProximityBleedDistance.val)
					{
						newSpatialBlend = (ProximityBleed.val / (ProximityBleedDistance.val - 1f) * (ProximityBleedDistance.val - sourceDistance));
						newSpatialBlend = 1 - (newSpatialBlend / 100);
					}
				}
			}

			// ******************
			// Setting the values
			// ******************
			// cutoff
			if( CASLowPassFilter != null ) CASLowPassFilter.cutoffFrequency = Mathf.Lerp(CASLowPassFilter.cutoffFrequency, newCutoff, OcclusionInterpolationSpeed.val * Time.deltaTime);
			if (VAMMoanPlugin != null)
			{
				float VAMMLPFVal = Mathf.Lerp(VAMMoanPlugin.GetFloatParamValue("VAMM Global Occlusion LPF"), newCutoff, OcclusionInterpolationSpeed.val * Time.deltaTime);				
				VAMMoanPlugin.SetFloatParamValue("VAMM Global Occlusion LPF", VAMMLPFVal);
			}
			if (CurrentAudioSourceControl != null)
			{
				// volume
				CurrentAudioSourceControl.SetFloatParamValue("volume", Mathf.Lerp( CurrentAudioSourceControl.GetFloatParamValue("volume"), newVolumeAtt, OcclusionInterpolationSpeed.val * Time.deltaTime) );
				// spatialBlend
				CurrentAudioSourceControl.SetFloatParamValue("spatialBlend", Mathf.Lerp( CurrentAudioSourceControl.GetFloatParamValue("spatialBlend"), newSpatialBlend, OcclusionInterpolationSpeed.val * Time.deltaTime) );
			}
			else if ( CurrentAudioSource != null )
			{
				// volume
				CurrentAudioSource.volume = Mathf.Lerp( CurrentAudioSource.volume, newVolumeAtt, OcclusionInterpolationSpeed.val * Time.deltaTime);
				// spatialBlend
				CurrentAudioSource.spatialBlend = Mathf.Lerp( CurrentAudioSource.spatialBlend, newSpatialBlend, OcclusionInterpolationSpeed.val * Time.deltaTime);
			}
		}
		
		private Vector3 CalculatePoint(Vector3 a, Vector3 b, float m, bool posOrneg, float vertOffset=0f)
		{
			float x;
			float z;
			float n = Vector3.Distance(new Vector3(a.x, 0f, a.z), new Vector3(b.x, 0f, b.z));
			float mn = (m / n);
			if (posOrneg)
			{
				x = a.x + (mn * (a.z - b.z));
				z = a.z - (mn * (a.x - b.x));
			}
			else
			{
				x = a.x - (mn * (a.z - b.z));
				z = a.z + (mn * (a.x - b.x));
			}
			return new Vector3(x, a.y + vertOffset, z);
		}
		
		private void CastLine(Vector3 Start, Vector3 End)
		{
			RaycastHit hit;
			RaycastHit hitBack;
			// Seems like the layer "0" is on what mostly anything that should hit is on (geo, meshes, CUAs...)
			// Based on Unity's doc, our layer mask is 1 << layerID, so 1 << 0 = 1
			Physics.Linecast(Start, End, out hit, 1);
			raycastTotal++;
			// Do we have a hit ?
			if ( hit.collider )
			{
				raycastHitCount++;
				showDebugLine(Start, End, Color.red);
				
				// Doing the backtrace only in Complex mode
				if (OcclusionType.val == "Complex")
				{
					Physics.Linecast(End, Start, out hitBack, 1);
					float thickness = sourceDistance - hit.distance - hitBack.distance;
					hitThicknessTotal += thickness;
				}
				
			}
			else
				showDebugLine(Start, End, Color.green);
		}
		
		// Thank you Sally Whitemane, bench based on SallySoundFX code
		// https://hub.virtamate.com/resources/sallysoundfx.15435/
		protected void StartBenchMark()
		{
			benchStopWatch.Reset();
			benchStopWatch.Start();
		}
		protected void StopBenchmark()
		{
			benchStopWatch.Stop();
			benchAvgUpdateTimes[benchAvgUpdateTimes.Length - 1] = benchStopWatch.Elapsed.TotalMilliseconds;
			Array.Copy(benchAvgUpdateTimes, 1, benchAvgUpdateTimes, 0, benchAvgUpdateTimes.Length - 1); // shift array
			benchAvgUpdateTime = 0;
			for (int i = 0; i < benchAvgUpdateTimes.Length; i++)
				benchAvgUpdateTime += benchAvgUpdateTimes[i];
			benchAvgUpdateTime = benchAvgUpdateTime / benchAvgUpdateTimes.Length;
		}
		
		// State of the plugin (displayed in the UI)
		private void UpdateStateInfos() {
			if (ShowStatistics.val == true)
			{
				if (internalTimer > nextStateUpdate)
				{
					string stateVal = "<color=#000><size=30><b>Statistics:</b></size></color>\n<color=#333><size=25>";
					stateVal += "<b>Occlusion time: " + benchAvgUpdateTime + "ms</b>\n";
					stateVal += "<b>Occlusion cutoff:</b> " + newCutoff + "\n";
					stateVal += "<b>Occlusion volume:</b> " + newVolumeAtt + "\n";
					stateVal += "<b>Occlusion spatial blend (bleed):</b> " + newSpatialBlend + "\n";
					if (OcclusionType.val == "Complex")
					{
						stateVal += "<b>Occlusion Thickness:</b> " +
						            (raycastHitCount > 0 ? hitThicknessTotal / raycastHitCount : 0) + "\n";
					}

					stateVal += "</size></color>";
					stateInfos.val = stateVal;

					nextStateUpdate = internalTimer + stateUpdateTime;
				}
			}
		}

		// **************************
		// Callbacks & similar
		// **************************
		protected void onUpdateDebugDisplay( bool activeState )
		{
			DebugEnabled = activeState;
			for (int i = 0; i < maxDebugLines; i++)
			{
				_debugLinesList[i].gameObject.SetActive(activeState);
			}
		}
		
		private void onChangeOcclusionType()
		{
			hitThicknessTotal = 0;
			onChangeOcclusionThicknessContribution();
			
			// Updating the UI
			Vector3 fastVecVal = new Vector3(1,1,1);
			Vector3 compVecVal = new Vector3(0,0,0);
			bool fastLayoutVal = false;
			bool compLayoutVal = true;
			if (OcclusionType.val == "Complex")
			{
				fastVecVal = new Vector3(0,0,0);
				compVecVal = new Vector3(1,1,1);
				fastLayoutVal = true;
				compLayoutVal = false;
			}
			
			// Iteration over UI elements
			foreach (UIDynamic uidElem in fastTypeUIElements)
			{
				uidElem.GetComponent<LayoutElement>().transform.localScale = fastVecVal;
				uidElem.GetComponent<LayoutElement>().ignoreLayout = fastLayoutVal;
			}
				
			foreach (UIDynamic uidElem in complexTypeUIElements)
			{
				uidElem.GetComponent<LayoutElement>().transform.localScale = compVecVal;
				uidElem.GetComponent<LayoutElement>().ignoreLayout = compLayoutVal;
			}
		}

		private void onChangeOcclusionThicknessContribution()
		{
			thicknessContribution = OcclusionThicknessContribution.val / 100;
			hitContribution = 1f - thicknessContribution;
		}
		
		private void onShowStatistics()
		{
			if (HelpUIElement != null)
			{
				if( ShowStatistics.val == false ) {			
					HelpUIElement.GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					HelpUIElement.GetComponent<LayoutElement>().ignoreLayout = true;
				} else {
					HelpUIElement.GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
					HelpUIElement.GetComponent<LayoutElement>().ignoreLayout = false;
				}
			}
		}

		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
        // Recursive function searching for a transform
        private Transform getObjectNamed( Transform parent, string searchName ) {
	        foreach( Transform child in parent )
	        {
		        if( child.name == searchName ) {
			        return child;
		        } else {
			        Transform childSearch = getObjectNamed(child,searchName);
			        if (childSearch != null) return childSearch;
		        }
	        }
			
	        return null;
        }
        
        // Thanks : https://forum.unity.com/threads/re-map-a-number-from-one-range-to-another.119437/
        float Remap(float source, float sourceFrom, float sourceTo, float targetFrom, float targetTo)
        {
	        return targetFrom + (source-sourceFrom)*(targetTo-targetFrom)/(sourceTo-sourceFrom);
        }
		
		// **************************
		// COROUTINES
		// **************************
		
		
		// **************************
		// Time to cleanup !
		// **************************
		void CleanupScene()
		{
			SoundSourceInitialized = false;
			DebugEnabled = false;
			
			// Removing components
			if (CASLowPassFilter != null)
			{
				Destroy(CASLowPassFilter);
			}
			
			// Resetting VAMMoan
			if (VAMMoanPlugin != null)
			{
				VAMMoanPlugin.SetFloatParamValue("VAMM Global Occlusion LPF", 22000.0f);
			}
			
			// Set Back Audio Source to the wanted value
			
			// Cleaning debug lines
			if (_debugLinesList != null)
			{
				for (int i = 0; i < maxDebugLines; i++)
				{
					Destroy(_debugLinesList[i].gameObject);
				}
			}
		}

		void OnEnable() {
			// Disabling debug
		}
		
		void OnDisable() {
			// Disabling debug

		}
		
		void OnDestroy() {			
			// Disabling debug
			CleanupScene();
		}

		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
    }

}
