using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMAtmosphere - Thunder v1.0
//
// A thunder generator

namespace VAMAThunderPlugin
{
    public class VAMAThunder : MVRScript
    {	
        public static string PLUGIN_PATH;
        public static string AUDIOBUNDLE_NAME = "vamatmosphere_thunder.audiobundle";


		public Transform CUARoot;
		public Transform ThunderAudioRoot;
		public Transform ThunderLightRoot;
		private GameObject GoTDebugAudio;
		private Atom AtLightFlash;
		public AudioSource AsThunder;
		
		private bool cuaInitialized = false;
		
		// VAM UI
		public JSONStorableBool ThunderEnabled;
		
		public JSONStorableFloat MinPauseThunder;
		public JSONStorableFloat MaxPauseThunder;
		
		public JSONStorableStringChooser LightFlash;
		public JSONStorableFloat MinFlashDuration;
		public JSONStorableFloat MaxFlashDuration;
		public JSONStorableFloat DoubleFlashProbability;

		public JSONStorableFloat ThunderVolume;
		public JSONStorableFloat ThunderPitch;
		public JSONStorableFloat ThunderSpatialization;
		public JSONStorableFloat ThunderReverb;
		
		public JSONStorableBool ShowDebug;
					
		// Security
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		
		// Assetbundle
		private Request audioBundleRequest = null;
		
		// Others
		private bool _thunderEnabled = true;
		private bool thunderRoutineEnabled = false;
		
		private List<string> thunderclapPaths;
		private List<AudioClip> thunderclapList;
		
		private List<string> sceneLightChoices;
		private List<Atom> sceneLightsList;
		private Atom currentLightFlash;
		
		// Actions
		public JSONStorableAction A_EnableThunder;
		public JSONStorableAction A_DisableThunder;
		public JSONStorableAction A_TriggerThunder;
		public JSONStorableAction A_TriggerThunderDouble;
		
			
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add VAMAThunder on a Custom Unity Asset Atom");
                    return;
                }
				
				PLUGIN_PATH = GetPluginPath(this);			
				
				ThunderEnabled = new JSONStorableBool("Thunder enabled", false, ThunderEnabledCallback);
				CreateToggle(ThunderEnabled, false);
				
				MinPauseThunder = new JSONStorableFloat("Pause between thunderclaps (min)", 10f, MinPauseThunderCallback, 1f, 60f);
				CreateSlider(MinPauseThunder, false);
				
				MaxPauseThunder = new JSONStorableFloat("Pause between thunderclaps (max)", 20f, MaxPauseThunderCallback, 1f, 60f);
				CreateSlider(MaxPauseThunder, false);
				
				LightFlash = new JSONStorableStringChooser("Light flash", sceneLightChoices, "", "Light flash", LightFlashCallback);
				UIDynamicPopup FAdp = CreateScrollablePopup(LightFlash, false);
				FAdp.labelWidth = 180f;
				
				MinFlashDuration = new JSONStorableFloat("Flash duration (min)", 0.05f, MinFlashDurationCallback, 0.01f, 2f);
				CreateSlider(MinFlashDuration, false);
				
				MaxFlashDuration = new JSONStorableFloat("Flash duration (max)", 0.2f, MaxFlashDurationCallback, 0.01f, 2f);
				CreateSlider(MaxFlashDuration, false);
				
				DoubleFlashProbability = new JSONStorableFloat("Double flash probability", 0.2f, DoubleFlashProbabilityCallback, 0f, 1f);
				CreateSlider(DoubleFlashProbability, false);
				
				ThunderVolume = new JSONStorableFloat("Thunder volume", 1f, ThunderVolumeCallback, 0f, 1f);
				CreateSlider(ThunderVolume, false);

				ThunderPitch = new JSONStorableFloat("Thunder pitch", 1f, ThunderPitchCallback, -3f, 3f);
				CreateSlider(ThunderPitch, false);
				
				ThunderSpatialization = new JSONStorableFloat("Thunder spatialization", 1f, ThunderSpatializationCallback, 0f, 1f);
				CreateSlider(ThunderSpatialization, false);
				
				ThunderReverb = new JSONStorableFloat("Thunder reverb", 1f, ThunderReverbCallback, 0f, 1.1f);
				CreateSlider(ThunderReverb, false);
							
				ShowDebug = new JSONStorableBool("Show bounds", false, ShowDebugCallback);
				CreateToggle(ShowDebug, false);
				
				/*UIDynamicButton stopLoop = CreateButton("Stop loop", true);
				if (stopLoop != null) {
					stopLoop.button.onClick.AddListener( () => { _thunderEnabled = false; } );
				}*/
			
				
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Help - Thunder</b></size></color>\n\n" + 
					"<color=#333>" +
					"<b>IMPORTANT:</b> If you switch asset in the Asset tab of the CUA, don't forget to reload the script to enable customization.\n\n" +
					"<b>Thunder enabled:</b> if the thunder is enabled at start. It will not change status if you use triggers to disable the thunder. This is the default value for the scene start.\n\n" +
					"<b>Pause between thunderclaps:</b> the minimum and maximum time between thunderclaps.\n\n" +
					"<b>Light flash:</b> the light used to make the lighting flash. It is not mandatory. Pixel lights recommended.\n\n" +
					"<b>Flash duration:</b> the minimum and maximum duration of the flashes (if you have selected a light above).\n\n" +
					"<b>Double flash probability:</b> the amount of chance the thunder will flash twice. 0 is never, 1 is everytime.\n\n" +
					"<b>Thunder volume:</b> the volume of the thunderclaps.\n\n" +
					"<b>Thunder pitch:</b> the pitch of the thunderclaps.\n\n" +
					"<b>Thunder spatialization:</b> if the sound of the thunder is spatialized or not.\n\n" +
					"<b>Thunder reverb:</b> the amount of signal sent to the reverb zone around.\n\n" +
					"<b>Show debug:</b> enable visual help to show the thunder position. Useful when you want to tweak the spatialization of the sound.\n\n" +
					"</color>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 1150.0f;
			
				RegisterBool(ThunderEnabled);
				
				RegisterFloat(MinPauseThunder);
				RegisterFloat(MaxPauseThunder);
				
				RegisterStringChooser(LightFlash);
				RegisterFloat(MinFlashDuration);
				RegisterFloat(MaxFlashDuration);
				RegisterFloat(DoubleFlashProbability);
				
				RegisterFloat(ThunderVolume);
				RegisterFloat(ThunderPitch);
				RegisterFloat(ThunderSpatialization);
				RegisterFloat(ThunderReverb);
				
				// Actions
				A_EnableThunder = new JSONStorableAction("Enable thunder", () => { ThunderEnabledCallback( true ); });
				A_DisableThunder = new JSONStorableAction("Disable thunder", () => { ThunderEnabledCallback( false ); });
				A_TriggerThunder = new JSONStorableAction("Trigger thunder", () => { TriggerThunderManual(); });
				A_TriggerThunderDouble = new JSONStorableAction("Trigger thunder (double flash forced)", () => { TriggerThunderDoubleManual(); });
				
				RegisterAction(A_EnableThunder);
				RegisterAction(A_DisableThunder);
				RegisterAction(A_TriggerThunder);
				RegisterAction(A_TriggerThunderDouble);
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAThunder - Exception caught: " + e);
            }
		}
		
		void Start()
        {

        }
		
		void Update() {
			// If we haven't made over 1min of init tests, we aren't initialized and the particle system (CUA) is still compatible, try to initialize
			if( initLoops < maxInitLoops && cuaInitialized == false ) {
				initLoops++;
				InitCUA();
			}
			
			if( initLoops == maxInitLoops ) {
				logError("VAMAThunder : CUA could not be initialized properly after 3600 tries. Please select a compatible CUA and reload the script.");
				initLoops++; // To avoid spamming the log
			}
		}


		// **************************
		// Functions
		// **************************
				
		private bool InitCUA() {
			if( cuaInitialized == true ) return true;
			
			CUARoot = getCUARoot( containingAtom.reParentObject );
			if( CUARoot == null ) {
				return false;
			}
			
			ThunderAudioRoot = getObjectNamed( CUARoot, "thunder_audio");
			if( ThunderAudioRoot == null ) {
				return false;
			}
						
			AsThunder = ThunderAudioRoot.gameObject.GetComponent<AudioSource>();
			if( AsThunder != null) {
				
				Transform TrDebug = getObjectNamed( CUARoot, "debug_audio");
				GoTDebugAudio = TrDebug.gameObject;
				
				cuaInitialized = true;

				// Loading the bundle
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + "/" + AUDIOBUNDLE_NAME, callback = OnAudioBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
			}
			return false;

		}
		
		protected void InitPaths() {
			// Creating assets paths list
			thunderclapPaths = new List<string>();
			thunderclapPaths.Add("Assets/VAMAtmosphere/audio_thunder/thunderclap_01.ogg");
			thunderclapPaths.Add("Assets/VAMAtmosphere/audio_thunder/thunderclap_02.ogg");
			thunderclapPaths.Add("Assets/VAMAtmosphere/audio_thunder/thunderclap_03.ogg");
			thunderclapPaths.Add("Assets/VAMAtmosphere/audio_thunder/thunderclap_04.ogg");
			thunderclapPaths.Add("Assets/VAMAtmosphere/audio_thunder/thunderclap_05.ogg");
			thunderclapPaths.Add("Assets/VAMAtmosphere/audio_thunder/thunderclap_06.ogg");
			thunderclapPaths.Add("Assets/VAMAtmosphere/audio_thunder/thunderclap_07.ogg");
		}
		
		private void UpdateThunderSettings() {
			ThunderEnabledCallback( ThunderEnabled.val );
			LightFlashCallback( LightFlash.val );
			
			ThunderVolumeCallback(ThunderVolume.val);
			ThunderPitchCallback(ThunderPitch.val);
			ThunderSpatializationCallback(ThunderSpatialization.val);
			ThunderReverbCallback(ThunderReverb.val);
			
			ShowDebugCallback( ShowDebug.val );
		}
		
		protected void UpdateDebugDisplay() {
			if ( GoTDebugAudio != null ) {
				//GoADebugMaxDist.transform.localScale = new Vector3( AmbienceMaxDist.val, AmbienceMaxDist.val, AmbienceMaxDist.val );
			}
		}
		
		protected void toggleLightFlash(bool enabled) {
			if( currentLightFlash != null ) {
				currentLightFlash.SetOn(enabled);
			}
		}

		protected void TriggerThunderManual() {
			AudioClip acThunder = getThunderAudioClip();
			AsThunder.PlayOneShot(acThunder);
			StartCoroutine(LoopThunderFlash(false));
		}
		
		protected void TriggerThunderDoubleManual() {
			AudioClip acThunder = getThunderAudioClip();
			AsThunder.PlayOneShot(acThunder);
			StartCoroutine(LoopThunderFlash(true));
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		private void OnAudioBundleLoaded(Request aRequest) {
			try
            {
				InitPaths();
				audioBundleRequest = aRequest;
				
				thunderclapList = new List<AudioClip>();
				foreach(string acPath in thunderclapPaths) {
					AudioClip ac = audioBundleRequest.assetBundle.LoadAsset<AudioClip>(acPath);
					thunderclapList.Add(ac);
				}

				sceneLightsList = new List<Atom>();
				sceneLightChoices = new List<string>();
				foreach(Atom at in SuperController.singleton.GetAtoms() ) {
					if( at.type == "InvisibleLight" ) {
						sceneLightsList.Add( at );
						sceneLightChoices.Add( at.name );
					}
				}
				LightFlash.choices = sceneLightChoices;
				
				UpdateThunderSettings();
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAThunder - Exception caught: " + e);
            }
		}
		
		protected void ThunderEnabledCallback(bool enabled) {
			_thunderEnabled = enabled;
			if( enabled == true && thunderRoutineEnabled == false ) {
				StartCoroutine("LoopThunder");
			} else if( enabled == false && thunderRoutineEnabled == true ) {
				StopCoroutine("LoopThunder");
				thunderRoutineEnabled = false;
			}
		}
		
		protected void MinPauseThunderCallback(float value) {}
		protected void MaxPauseThunderCallback(float value) {}
		
		protected void LightFlashCallback(string choice) {
			if( sceneLightsList != null && sceneLightChoices != null && sceneLightsList.Count > 0 && sceneLightChoices.Count > 0 && choice != "" ) {
				currentLightFlash = sceneLightsList[ sceneLightChoices.IndexOf(choice) ];
				currentLightFlash.SetOn(false);
			}
		}
		protected void MinFlashDurationCallback(float value) {}
		protected void MaxFlashDurationCallback(float value) {}
		protected void DoubleFlashProbabilityCallback(float value) {}

		protected void ThunderVolumeCallback(float volume) {
			if( CUARoot && AsThunder ) {
				AsThunder.volume = volume;
			}
		}

		protected void ThunderPitchCallback(float pitch) {
			if( CUARoot && AsThunder ) {
				AsThunder.pitch = pitch;
			}
		}
		
		protected void ThunderSpatializationCallback(float val) {
			if( CUARoot && AsThunder ) {
				AsThunder.spatialBlend = val;
			}
		}
		
		protected void ThunderReverbCallback(float val) {
			if( CUARoot && AsThunder ) {
				AsThunder.reverbZoneMix = val;
			}
		}
		
		protected void ShowDebugCallback(bool enabled) {
			if( CUARoot ) {				
				if( enabled == true ) {
					GoTDebugAudio.SetActive(true);
					
					UpdateDebugDisplay();
				} else {
					GoTDebugAudio.SetActive(false);
					
					UpdateDebugDisplay();
				}
			}
		}
		
		private AudioClip getThunderAudioClip() {
			int mIndex = UnityEngine.Random.Range(0, thunderclapList.Count);
			return thunderclapList[mIndex];
		}

		// **************************
		// COROUTINES
		// **************************
		private IEnumerator LoopThunder() 
		{
			thunderRoutineEnabled = true;
			
			// Initial pause
			float firstPauseDuration = UnityEngine.Random.Range(MinPauseThunder.val, MaxPauseThunder.val);
			yield return new WaitForSeconds(firstPauseDuration);

			
			while (_thunderEnabled == true)
			{
				AudioClip acThunder = getThunderAudioClip();
				AsThunder.clip = acThunder;
				float clipDuration = acThunder.length;
				AsThunder.Play();
				StartCoroutine(LoopThunderFlash(false));
				
				float pauseDuration = clipDuration + UnityEngine.Random.Range(MinPauseThunder.val, MaxPauseThunder.val);
				yield return new WaitForSeconds(pauseDuration);

			}

			thunderRoutineEnabled = false;
		}
			
		private IEnumerator LoopThunderFlash( bool forceDoubleFlash ) 
		{
			float lightPauseDuration = UnityEngine.Random.Range(MinFlashDuration.val, MaxFlashDuration.val);
			toggleLightFlash(true);
			yield return new WaitForSeconds(lightPauseDuration);
			toggleLightFlash(false);
			
			// Do we trigger a second flash ?
			float doDoubleFlash = UnityEngine.Random.Range(0f, 1f);
			if( doDoubleFlash <= DoubleFlashProbability.val || forceDoubleFlash == true ) {
				// Pause before the second flash
				float flashPauseDuration = UnityEngine.Random.Range(MinFlashDuration.val, MaxFlashDuration.val);
				yield return new WaitForSeconds(flashPauseDuration);
				
				// Second flash
				lightPauseDuration = UnityEngine.Random.Range(0.07f, 0.11f);
				toggleLightFlash(true);
				yield return new WaitForSeconds(lightPauseDuration);
				toggleLightFlash(false);
			}
		}

		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
        // Recursive function searching for a transform
        private Transform getCUARoot( Transform parent ) {
            foreach( Transform child in parent )
            {
				if( child.name.Contains("VAMAThunder") ) {
					return child;
				} else {
					Transform childSearch = getCUARoot(child);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		private Transform getObjectNamed( Transform parent, string objName ) {
            foreach( Transform child in parent )
            {
				if( child.name == objName ) {
					return child;
				} else {
					Transform childSearch = getObjectNamed(child,objName);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnEnable() {
			// Disabling debug
			ShowDebugCallback( ShowDebug.val );
			
		}
		
		void OnDisable() {
			// Disabling debug
			ShowDebugCallback( false );
		}
		
		void OnDestroy() {
			// Disabling debug
			ShowDebugCallback( false );
			
			// Cleaning modified native VAM assets
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + "/" + AUDIOBUNDLE_NAME);

		}

		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		

	}
}
